/*
     This file is part of GNUnet.
     (C) 2001, 2002 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * GNUNET MAIN. This is the server-process that must run on each
 * and every GNUNET node. The code in here initializes the subsystems
 * and then dispatches incomming UDP packets using multiple threads
 * to the handler(s).
 *
 * @file server/gunnet.c
 * @author Christian Grothoff
 * @author Larry Waldo
 * @author Tzvetan Horozov
 **/

#include "config.h"
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/stat.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <pthread.h>
#include <setjmp.h>
 
#include "configuration.h"
#include "util/ipcheck.h"
#include "util/semaphore.h"
#include "util/statuscalls.h"
#include "util/contentdatabase.h"
#include "util/logging.h"
#include "server/startup.h"
#include "server/handler.h"
#include "server/pingpong.h"
#include "server/keyservice.h" 
#include "server/heloexchange.h"
#include "server/trustservice.h"
#include "server/udpserver.h"
#include "knownhosts.h"
#include "lookup.h"
#include "statistics.h"

static sigjmp_buf jmpbuf;
static volatile sig_atomic_t canjump;

#if PRINT_CRONDOT
static void printDot(void * unused) {
  print(".");
}
#endif

/**
 * Try a propper shutdown of gnunetd (experimental...).
 **/
static void shutdown_gnunetd(int signum) {
  if (canjump == 0) {
    WARNING("Can not shutdown, siglongjump not possible, canjump == 0!\n");
    return;
  }
  canjump = 0;
  siglongjmp(jmpbuf, 1); /* jumps into main! see also Stevens "Advanced Programming in the Unix Environment", page 300, 301 */
}

static void reread_config_helper(void * unused) {
  MESSAGE("re-reading configuration\n");
  updateConfiguration();
  setLogLevel(getLogLevel());
  MESSAGE("re-reading configuration done\n");
}

/**
 * Re-read the configuration file. 
 **/
static void reread_config(int signum) {
  DEBUG("Received SIGHUP %d\n",signum);
  addCronJob(&reread_config_helper,
	     1, 0, NULL);
}

/**
 * The main method of GNUnet. And here is how it works:
 * 1) initialize subsystems 
 *    (configuration, semaphores, key library, output queue)
 * 2) initialize synchronization mechanism for the threads
 * 3) initialize threads, they execute "threadMain"
 *    and query the input queue with incomming packets
 * 4) initialize network (UDP listen)
 * 5) forever: get UDP packet, put it in the queue for the threads
 * 6) [not yet reachable]: close socket on exit
 **/
int main(int argc, char * argv[]) {
  int filedes[2]; /* pipe between client and parent */
  FileName configFile = DEFAULT_CONFIG_FILE;
  int nullfd;

  /* initialize subsystems */
  if (parseCommandLine(argc, argv, (char**)&configFile))
    return 0; /* parse error */

  initConfiguration(configFile);
  checkCompiler();

  /* Don't hold the wrong FS mounted */
  if (chdir("/") < 0) {
    perror("chdir");
    return 1;
  }
  umask(0);
  if (NO == debug_flag) {
    pid_t pid;

    pipe(filedes);
    pid = fork();
    if (pid < 0) {
      perror("fork");
      return 1;
    }
    if (pid) {  /* Parent */
      int ok;
      char c;

      close(filedes[1]); /* we only read */
      ok = SYSERR;
      while (0 < read(filedes[0], &c, sizeof(char))) {
	if (c == '.') 
	  ok = OK;
      }      
      /* wait here for child error/success status */
      /* print pid of child for /var/run/gnunetd.pid */
      /* FIXME (?): printf("%u\n", (unsigned)pid); */
      fflush(stdout);
      if (ok == OK)
	exit(0);       /* Avoid atexit() problems */
      else
	exit(1); /* child reported error */
    }
    close(filedes[0]); /* we only write */
    nullfd = open("/dev/null", O_CREAT | O_RDWR | O_APPEND);
    if (nullfd < 0) {
      perror("/dev/null");
      return 1;
    }   
    /* child - close fds linking to invoking terminal, but
     * close usual incoming fds, but redirect them somewhere
     * useful so the fds don't get reallocated elsewhere.
     */
    if (dup2(nullfd,0) < 0 || 
	dup2(nullfd,1) < 0 || 
	dup2(nullfd,2) < 0) {
      perror("dup2"); /* Should never happen */
      return 1;
    }
    pid = setsid(); /* Detach from controlling terminal */
    initLogging(getLogFilename(),
		getLogLevel(),
		YES);
  } else
    initLogging(NULL, /* send to stderr */
		getLogLevel(),
		YES);


  initStatistics();
  initContentDatabase(getContentDirectory());
  initCron();
  initTrustService();
  initTCPIO();
  initStatusCalls(getNetInterfaces(),
          getMaxUpstreamBPSTotal(),
          getMaxDownstreamBPSTotal());
  initAddress();
  initKnownhosts();
  initKeyService();
  initPingPong();
  initUDPServer();
  initConnection(); 
  initLookup();
  fullInitLookup();
  initRouting();
  startTCPListener();
  initHeloExchange();

  /* start control thread */
#if PRINT_CRONDOT
  addCronJob(&printDot,
	     1, 1, NULL);
#endif  
  startCron();

  /* initialize signal handler (CTRL-C / SIGTERM) */
  canjump = 0;
  {
    struct sigaction sig;
    struct sigaction oldsig;
    sig.sa_handler = &shutdown_gnunetd;
    sigemptyset(&sig.sa_mask);
#ifdef SA_INTERRUPT
    sig.sa_flags = SA_RESETHAND | SA_INTERRUPT; /* SunOS */
#else
    sig.sa_flags = SA_RESETHAND | SA_RESTART;
#endif    
    sigaction(SIGINT,  &sig, &oldsig);
    sigaction(SIGTERM, &sig, &oldsig);
    sigaction(SIGQUIT, &sig, &oldsig);

    sig.sa_handler = &reread_config;
#ifdef SA_INTERRUPT
     sig.sa_flags = SA_INTERRUPT; /* SunOS */
#else
     sig.sa_flags = SA_RESTART;
#endif    
    sigaction(SIGHUP, &sig, &oldsig);
  }

  /* send success to parent (by closing stream and never writing to it) */
  if (NO == debug_flag) {
    char c = '.';
    write(filedes[1], &c, sizeof(char)); /* signal success */
    close(filedes[1]);
  }
  {
    FILE * pidfd;
    
    pidfd = fopen(getPIDFile(), "w");
    if (pidfd == NULL) {
      WARNING("Could not write PID to file %s\n", getPIDFile());
    } else {
      fprintf(pidfd, "%u", getpid());
      fclose(pidfd);
    }
  }

  if (sigsetjmp(jmpbuf, 1)) {
    MESSAGE("GNUnetd shutting down\n");
    stopCron();
#if PRINT_CRONDOT
    delCronJob(&printDot,
	       1, NULL);
#endif  
    DEBUG("Cron stopped!\n");
    doneHeloExchange();  
    doneTCPListener();
    doneUDPServer();
    DEBUG("Closing connections\n");
    doneConnection();
    DEBUG("Closing database.\n");
    doneContentDatabase();
    doneLookup();
    donePingPong();
    doneKeyService();
    doneKnownhosts();
    doneTrustService();
    doneAddress();
    doneTCPIO();
    doneStatusCalls();
    doneCron();
    unlink(getPIDFile());
    doneConfiguration();
    doneStatistics();
    MESSAGE("GNUnetd shutdown complete\n"); /* here because afterwards logging is disabled! */
    doneLogging();
    return 0;
  } else {
    canjump = 1;
    runUDPDispatch();
    /* see shutdown_gnunetd - this can never happen ... */
    return 0;
  }
} 

/* You have reached the end of GNUnet. You can shutdown your
   computer and get a live now. */


 
